#!/usr/bin/env python3
#
# wireguard4netns
#
# Copyright (c) 2022 Carnegie Mellon University
# SPDX-License-Identifier: MIT
#

import os
import shutil
import subprocess
import sys
import tarfile
import urllib.request
from pathlib import Path
from tempfile import TemporaryDirectory

GOLANG = "https://storage.googleapis.com/golang/go{}.linux-amd64.tar.gz"
GOLANG_VERSION = "1.19.4"


def build(*_setup_kwargs):
    """Build wireguard-go"""
    dest = Path("src") / "wireguard4netns" / "wireguard-go"

    if dest.exists():
        print(f"{dest} already exists, skipping build")
        return

    go = shutil.which("go")
    patch = shutil.which("patch")

    # copy wireguard-go source into a temporary directory, patch and compile
    with TemporaryDirectory() as tmp:
        if go is None:
            # fetch + extract go compiler
            print(f"Downloading golang-v{GOLANG_VERSION}")
            stream = urllib.request.urlopen(GOLANG.format(GOLANG_VERSION))
            golang = tarfile.open(fileobj=stream, mode="r|gz")
            golang.extractall(path=tmp)
            go = str(Path(tmp, "go", "bin", "go").resolve())

        print("Copying wireguard-go source")
        tmp_root = Path(tmp, "wireguard-go")
        shutil.copytree("wireguard-go", tmp_root)

        # when building from a git checkout this ends up pointing at a
        # non-existent path and we don't need it, so just clean it up.
        gitlink = tmp_root.joinpath(".git")
        if gitlink.exists():
            gitlink.unlink()

        # apply patch to disable SetMTU()/MTU() because wireguard-go won't be
        # able to see the tuntap interface in the unprivileged network namespace
        print("Patching wireguard-go source")
        patchfile = Path("wireguard-go.patch")
        subprocess.run([patch, "-p1"], stdin=patchfile.open(), cwd=tmp_root, check=True)

        # to avoid downloading already cached modules
        GOMODCACHE = subprocess.run(
            [go, "env", "GOMODCACHE"], stdout=subprocess.PIPE, check=True
        ).stdout
        env = dict(os.environ, GOPATH=str(Path(tmp).resolve()), GOMODCACHE=GOMODCACHE)

        print("Building custom wireguard-go binary")
        subprocess.run([go, "get", "-d"], cwd=tmp_root, env=env, check=True)
        subprocess.run(
            [
                go,
                "build",
                "-o",
                str(dest.resolve()),
                "-ldflags=-s -w -X golang.zx2c4.com/wireguard/ipc.socketDirectory=.",
            ],
            cwd=tmp_root,
            env=env,
            check=True,
        )


def check_submodule():
    # make sure wireguard-go submodule is initialized
    sys.exit(0 if Path("wireguard-go", "LICENSE").exists() else 1)


def clean_dist():
    """clean up leftover files from old builds"""
    dist = Path("dist")

    for sdist in dist.glob("*.tar.gz"):
        sdist.unlink()

    for wheel in dist.glob("*.whl"):
        wheel.unlink()


SCRIPT = """\
cd /tmp
curl {golang} --silent --location | tar -xz
export PATH="/tmp/go/bin:$PATH" HOME=/tmp
for py in {pythons}; do
    "/opt/python/$py/bin/pip" wheel --no-deps --wheel-dir /tmp /dist/*.tar.gz
done
ls *.whl | xargs -n1 --verbose auditwheel repair --wheel-dir /dist
ls -al /dist
"""


def wheels():
    """Build wheels from the pypa/manylinux2014_x86_64 Docker container

    Based on how setuptools-golang builds, but using a newer manylinux
    container and hardcoded golang/python versions.
    """
    golang = GOLANG.format(GOLANG_VERSION)
    pythons = [
        "cp37-cp37m",
        "cp38-cp38",
        "cp39-cp39",
        "cp310-cp310",
        "cp311-cp311",
    ]

    dist = Path("dist")

    subprocess.run(
        [
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{dist.resolve()}:/dist:rw",
            "--user",
            f"{os.getuid()}:{os.getgid()}",
            "quay.io/pypa/manylinux2014_x86_64:latest",
            "bash",
            "-o",
            "pipefail",
            "-euxc",
            SCRIPT.format(golang=golang, pythons=" ".join(pythons)),
        ],
        check=True,
    )


if __name__ == "__main__":
    build()



#
# wireguard4netns
#
# Copyright (c) 2022 Carnegie Mellon University
# SPDX-License-Identifier: MIT
#
"""Fork wireguard-go process and configure the tunnel"""

from __future__ import annotations

import os
import subprocess
import time
from contextlib import closing, contextmanager
from pathlib import Path
from typing import Iterator

import importlib_resources
from wireguard_tools import WireguardConfig
from wireguard_tools.wireguard_uapi import WireguardUAPIDevice

from .tundev import create_tun_in_netns


@contextmanager
def fork_wireguard_go(
    ns_pid: int, interface: str, tmpdir: Path
) -> Iterator[WireguardUAPIDevice]:

    # locate wireguard-go binary
    wireguard_go = importlib_resources.files("wireguard4netns").joinpath("wireguard-go")

    tundev = create_tun_in_netns(ns_pid, interface)
    try:
        subprocess.Popen(
            [wireguard_go, interface],
            env=dict(
                WG_PROCESS_FOREGROUND="1",
                WG_TUN_FD=str(tundev),
                # LOG_LEVEL="debug",
            ),
            pass_fds=(tundev,),
            cwd=tmpdir,
        )
    except FileNotFoundError:
        os.close(tundev)
        raise

    # wait for uapi socket
    uapi_path = Path(tmpdir, interface).with_suffix(".sock")
    while not uapi_path.exists():
        time.sleep(0.1)

    with closing(WireguardUAPIDevice(uapi_path)) as device:
        yield device

    # do we even care to wait for the child to exit?
    # no we don't, it will terminate when the device or tmpdir are cleaned up


def create_wireguard_tunnel(
    ns_pid: int, interface: str, wg_config: WireguardConfig, tmpdir: Path | None = None
) -> None:
    if tmpdir is None:
        tmpdir = Path()

    with fork_wireguard_go(ns_pid, interface, tmpdir) as wg_device:
        wg_device.set_config(wg_config)



#
# wireguard4netns
#
# Copyright (c) 2022 Carnegie Mellon University
# SPDX-License-Identifier: MIT
#
__version__ = "0.1.1"

from .wireguard_daemon import create_wireguard_tunnel

__all__ = [
    "create_wireguard_tunnel",
]



#
# wireguard4netns
#
# Copyright (c) 2022 Carnegie Mellon University
# SPDX-License-Identifier: MIT
#
"""Create a TUN device in an unprivileged network namespace"""

from __future__ import annotations

import ctypes
import ctypes.util
import errno
import fcntl
import os
import struct
from multiprocessing import Pipe, Process
from multiprocessing.connection import Connection
from multiprocessing.reduction import recv_handle, send_handle
from pathlib import Path

from pyroute2 import NDB
from typing_extensions import Literal

_libc = ctypes.CDLL(ctypes.util.find_library("c"), use_errno=True)


def setns(target_ns_pid: int, ns_type: Literal["user", "net"]) -> None:
    ns_path = Path("/proc", str(target_ns_pid), "ns", ns_type)
    with ns_path.open() as nsfd:
        if _libc.setns(nsfd.fileno(), 0) == -1:
            e = ctypes.get_errno()
            raise OSError(e, errno.errorcode[e])


def _create_tun_child(target_ns_pid: int, tundev_name: str, writer: Connection) -> None:
    # join target user and network namespaces
    setns(target_ns_pid, "user")
    setns(target_ns_pid, "net")

    tundev = os.open("/dev/net/tun", os.O_RDWR)
    assert tundev != -1

    TUNSETIFF = 0x400454CA
    IFF_TUN = 0x0001

    # create tuntap device
    ifreq = struct.pack("16sH22x", tundev_name.encode(), IFF_TUN)
    fcntl.ioctl(tundev, TUNSETIFF, ifreq)

    # pass tunnel device handle back to parent
    send_handle(writer, tundev, None)

    # set interface mtu
    with NDB() as ndb:
        ndb.interfaces[tundev_name].set("mtu", 1420).commit()


def create_tun_in_netns(target_ns_pid: int, tundev_name: str) -> int:
    reader, writer = Pipe()
    p = Process(target=_create_tun_child, args=(target_ns_pid, tundev_name, writer))
    p.start()
    tundev = recv_handle(reader)
    p.join()
    return tundev



#
# wireguard4netns
#
# Copyright (c) 2022 Carnegie Mellon University
# SPDX-License-Identifier: MIT
#

from __future__ import annotations

import argparse
from pathlib import Path

from wireguard_tools import WireguardConfig

from . import __version__
from .wireguard_daemon import create_wireguard_tunnel


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument("--tmpdir", type=Path, default=".")
    parser.add_argument("ns_pid", type=int)
    parser.add_argument("interface")
    parser.add_argument("config", metavar="wireguard.conf", type=argparse.FileType("r"))
    args = parser.parse_args()

    wg_config = WireguardConfig.from_wgconfig(args.config)
    create_wireguard_tunnel(args.ns_pid, args.interface, wg_config, args.tmpdir)
    return 0
